"""
TNReason: A package for reasoning with Tensor Networks with the subpackages
* engine: Provides the bindings to tensor network contractors
* encoding: Represents maps by tensors
* algorithms: Implements general purpose optimization algorithms on tensor networks
* knowledge: Executes deductive and inductive reasoning tasks based on markov logic networks
"""

__version__ = "1.0.0"
